@@ -1,15 +1,14 @@
 """ Data generators for training/inference with siamese Keras model.
 """
+import warnings
 from typing import List, Iterator, NamedTuple
+
 import numpy as np
 import pandas as pd
 from tensorflow.keras.utils import Sequence
 
 from .typing import BinnedSpectrumType
 
-# Set random seed for reproducibility
-np.random.seed(42)
-
 
 class SpectrumPair(NamedTuple):
     """
@@ -62,6 +61,14 @@ class DataGeneratorBase(Sequence):
             Change peak intensities by a random number between 0 and augment_intensity.
             Default=0.1, which means that intensities are multiplied by 1+- a random
             number within [0, 0.1].
+        augment_noise_max
+            Max number of 'new' noise peaks to add to the spectrum, between 0 to `augment_noise_max`
+            of peaks are added.
+        augment_noise_intensity
+            Intensity of the 'new' noise peaks to add to the spectrum
+        use_fixed_set
+            Toggles using a fixed dataset, if set to True the same dataset will be generated each
+            epoch. Default is False.
         """
 
         self._validate_labels(reference_scores_df)
@@ -73,6 +80,7 @@ class DataGeneratorBase(Sequence):
         self.reference_scores_df = self._transform_to_inchikey14(self.reference_scores_df)
         self._collect_and_validate_inchikeys()
         self.dim = dim
+        self.fixed_set = dict()
 
     def _collect_and_validate_inchikeys(self):
         """Collect all inchikeys14 (first 14 characters) of all binned_spectrums
@@ -136,6 +144,14 @@ class DataGeneratorBase(Sequence):
             Change peak intensities by a random number between 0 and augment_intensity.
             Default=0.1, which means that intensities are multiplied by 1+- a random
             number within [0, 0.1].
+        augment_noise_max
+            Max number of 'new' noise peaks to add to the spectrum, between 0 to `augment_noise_max`
+            of peaks are added.
+        augment_noise_intensity
+            Intensity of the 'new' noise peaks to add to the spectrum
+        use_fixed_set
+            Toggles using a fixed dataset, if set to True the same dataset will be generated each
+            epoch. Default is False.
         """
         defaults = dict(
             batch_size=32,
@@ -148,6 +164,7 @@ class DataGeneratorBase(Sequence):
             augment_intensity=0.4,
             augment_noise_max=10,
             augment_noise_intensity=0.01,
+            use_fixed_set=False
         )
 
         # Set default parameters or replace by **settings input
@@ -159,13 +176,17 @@ class DataGeneratorBase(Sequence):
                 settings[key] = defaults[key]
         assert 0.0 <= settings["augment_removal_max"] <= 1.0, "Expected value within [0,1]"
         assert 0.0 <= settings["augment_removal_intensity"] <= 1.0, "Expected value within [0,1]"
+        if settings["use_fixed_set"] and settings["shuffle"]:
+            warnings.warn('When using a fixed set, data will not be shuffled')
+        if settings["use_fixed_set"]:
+            np.random.seed(42)
         self.settings = settings
 
-    def _find_match_in_range(self, inchikey1, target_score_range, max_range=0.4):
+    def _find_match_in_range(self, inchikey1, target_score_range):
         """Randomly pick ID for a pair with inchikey_id1 that has a score in
         target_score_range. When no such score exists, iteratively widen the range
-        in steps of 0.1 until a max of max_range. If still no match is found take
-        a random ID.
+        in steps of 0.1.
+
         Parameters
         ----------
         inchikey1
@@ -178,7 +199,7 @@ class DataGeneratorBase(Sequence):
         extend_range = 0
         low, high = target_score_range
         inchikey2 = None
-        while extend_range < max_range:
+        while inchikey2 is None:
             matching_inchikeys = self.reference_scores_df.index[
                 (self.reference_scores_df[inchikey1] > low - extend_range)
                 & (self.reference_scores_df[inchikey1] <= high + extend_range)]
@@ -186,22 +207,23 @@ class DataGeneratorBase(Sequence):
                 matching_inchikeys = matching_inchikeys[matching_inchikeys != inchikey1]
             if len(matching_inchikeys) > 0:
                 inchikey2 = np.random.choice(matching_inchikeys)
-                break
             extend_range += 0.1
-
-        # Part 2 - if still no match is found take the inchikey that has highest similarity score
-        # with inchikey1
-        # TODO: Why are we taking the highest score here? Shouldn't we take the one that is closest
-        #  to the center of the bin?
-        if not inchikey2:
-            inchikey2 = self.reference_scores_df[inchikey1][self.reference_scores_df.index != inchikey1].idxmax()
-
         return inchikey2
 
     def __getitem__(self, batch_index: int):
-        """Generate one batch of data"""
+        """Generate one batch of data.
+
+        If use_fixed_set=True we try retrieving the batch from self.fixed_set (or store it if
+        this is the first epoch). This ensures a fixed set of data is generated each epoch.
+        """
+        if self.settings['use_fixed_set'] and batch_index in self.fixed_set:
+            return self.fixed_set[batch_index]
+        if self.settings['use_fixed_set'] and batch_index == 0:
+            np.random.seed(42)
         spectrum_pairs = self._spectrum_pair_generator(batch_index)
         X, y = self.__data_generation(spectrum_pairs)
+        if self.settings['use_fixed_set']:
+            self.fixed_set[batch_index] = (X, y)
         return X, y
 
     def _data_augmentation(self, spectrum_binned):
@@ -229,14 +251,22 @@ class DataGeneratorBase(Sequence):
         if self.settings["augment_intensity"]:
             # TODO: Factor out function with documentation + example?
             values = (1 - self.settings["augment_intensity"] * 2 * (np.random.random(values.shape) - 0.5)) * values
-        # Augmentation 3: Add noise peaks
+        # Augmentation 3: Peak addition
         if self.settings["augment_noise_max"] and self.settings["augment_noise_max"] > 0:
-            n_noise_peaks = np.random.randint(0, self.settings["augment_noise_max"])
-            idx_noise_peaks = np.random.randint(0, self.dim, n_noise_peaks)
-            idx_noise_peaks = idx_noise_peaks[np.isin(idx_noise_peaks, idx, invert=True)]
-            idx = np.concatenate((idx, idx_noise_peaks))
-            values = np.concatenate((values,
-                                     self.settings["augment_noise_intensity"] * np.random.random(len(idx_noise_peaks))))
+            idx, values = self._peak_addition(idx, values)
+        return idx, values
+
+    def _peak_addition(self, idx, values):
+        """
+        For each of between 0-augment_noise_max randomly selected zero-intensity bins
+        that bin’s intensity is set to random values between 0 and augment_noise_intensity
+        """
+        n_noise_peaks = np.random.randint(0, self.settings["augment_noise_max"])
+        idx_no_peaks = np.setdiff1d(np.arange(0, self.dim), idx)
+        idx_noise_peaks = np.random.choice(idx_no_peaks, n_noise_peaks)
+        idx = np.concatenate((idx, idx_noise_peaks))
+        new_values = self.settings["augment_noise_intensity"] * np.random.random(len(idx_noise_peaks))
+        values = np.concatenate((values, new_values))
         return idx, values
 
     def _get_spectrum_with_inchikey(self, inchikey: str) -> BinnedSpectrumType:
@@ -429,9 +459,11 @@ class DataGeneratorAllInchikeys(DataGeneratorBase):
         self.on_epoch_end()
 
     def __len__(self):
-        """Denotes the number of batches per epoch"""
-        # TODO: this means we don't see all data every epoch, because the last half-empty batch
-        #  is omitted. I guess that is expected behavior? --> Yes, with the shuffling in each epoch that seem OK to me (and makes the code easier).
+        """Denotes the number of batches per epoch
+        NB1: self.reference_scores_df only contains 'selected' inchikeys, see `self._data_selection`.
+        NB2: We don't see all data every epoch, because the last half-empty batch is omitted. 
+        This is expected behavior, with the shuffling this is OK.
+        """
         return int(self.settings["num_turns"]) * int(np.floor(len(self.reference_scores_df) / self.settings[
             "batch_size"]))
 
