@@ -179,7 +179,12 @@ class SiameseModel:
                            name='head')
 
     def _construct_from_keras_model(self, keras_model):
-        assert isinstance(keras_model, keras.Model), "Expected keras model as input."
+        def valid_keras_model(given_model):
+            assert given_model.layers, "Expected valid keras model as input."
+            assert len(given_model.layers) > 2, "Expected more layers"
+            assert len(keras_model.layers[2].layers) > 1, "Expected more layers for base model"
+            
+        valid_keras_model(keras_model)
         self.base = keras_model.layers[2]
         self.model = keras_model
 
